# -*- coding: utf-8 -*-

# DO NOT EDIT THIS FILE!
# This file has been autogenerated by dephell <3
# https://github.com/dephell/dephell

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import os.path

readme = ""
here = os.path.abspath(os.path.dirname(__file__))
readme_path = os.path.join(here, "README.rst")
if os.path.exists(readme_path):
    with open(readme_path, "rb") as stream:
        readme = stream.read().decode("utf8")

setup(
    long_description=readme,
    name="event_metrics",
    version="0.1.0",
    description="An embedded, event-time metric collection library",
    python_requires="==3.*,>=3.5.0",
    author="simon-mo",
    author_email="xmo@berkeley.edu",
    packages=["event_metrics"],
    package_dir={"": "."},
    package_data={},
    install_requires=["numpy"],
    extras_require={
        "dev": [
            "pytest==5.*,>=5.2.0",
            "pytest-benchmark==3.*,>=3.2.2",
            "pytest-sugar==0.*,>=0.9.2",
        ]
    },
)
